obj/player_types.o dep/player_types.d : src/player_types.cpp src/player_types.h src/main.h \
 src/debug.h src/globaltypes.h src/externs.h src/globaldata.h src/text.h \
 src/environment.h src/network.h src/gfxData.h
