obj/decor.o dep/decor.d : src/decor.cpp src/decor.h src/physobj.h src/globaltypes.h \
 src/virtobj.h src/main.h src/debug.h src/externs.h src/globaldata.h \
 src/text.h src/environment.h src/network.h src/gfxData.h \
 src/debris_pool.h src/sound.h src/tank.h src/floattext.h
