obj/optionscreens.o dep/optionscreens.d : src/optionscreens.cpp src/optionscreens.h src/menu.h \
 src/optionitem.h src/optionitembase.h src/environment.h src/main.h \
 src/debug.h src/globaltypes.h src/externs.h src/globaldata.h src/text.h \
 src/network.h src/gfxData.h src/optiontypes.h src/optionitemmenu.h \
 src/optionitemplayer.h src/button.h src/player.h src/player_types.h \
 src/files.h src/sound.h
