obj/optionitembase.o dep/optionitembase.d : src/optionitembase.cpp src/button.h src/main.h \
 src/debug.h src/globaltypes.h src/externs.h src/globaldata.h src/text.h \
 src/environment.h src/network.h src/gfxData.h src/menu.h \
 src/optionitem.h src/optionitembase.h src/optiontypes.h \
 src/optionitemmenu.h src/optionitemplayer.h src/floattext.h \
 src/virtobj.h
