obj/satellite.o dep/satellite.d : src/satellite.cpp src/environment.h src/main.h src/debug.h \
 src/globaltypes.h src/externs.h src/globaldata.h src/text.h \
 src/network.h src/gfxData.h src/satellite.h src/beam.h src/physobj.h \
 src/virtobj.h
