obj/optionitemplayer.o dep/optionitemplayer.d : src/optionitemplayer.cpp src/optionitemplayer.h \
 src/optionitembase.h src/environment.h src/main.h src/debug.h \
 src/globaltypes.h src/externs.h src/globaldata.h src/text.h \
 src/network.h src/gfxData.h src/optiontypes.h src/player.h \
 src/player_types.h src/floattext.h src/virtobj.h
