obj/menu.o dep/menu.d : src/menu.cpp src/optioncontent.h src/optiontypes.h \
 src/globaltypes.h src/optionitemcolour.h src/optionitembase.h \
 src/environment.h src/main.h src/debug.h src/externs.h src/globaldata.h \
 src/text.h src/network.h src/gfxData.h src/button.h src/menu.h \
 src/optionitem.h src/optionitemmenu.h src/optionitemplayer.h \
 src/player.h src/player_types.h src/clock.h
