obj/tank.o dep/tank.d : src/tank.cpp src/floattext.h src/main.h src/debug.h \
 src/globaltypes.h src/externs.h src/globaldata.h src/text.h \
 src/environment.h src/network.h src/gfxData.h src/virtobj.h \
 src/explosion.h src/physobj.h src/teleport.h src/missile.h src/player.h \
 src/player_types.h src/beam.h src/tank.h src/sound.h
