obj/atanks.o dep/atanks.d : src/atanks.cpp src/debug.h src/globals.h src/globaldata.h \
 src/main.h src/globaltypes.h src/externs.h src/environment.h \
 src/network.h src/gfxData.h src/text.h src/optionscreens.h src/menu.h \
 src/optionitem.h src/optionitembase.h src/optiontypes.h \
 src/optionitemmenu.h src/optionitemplayer.h src/button.h src/player.h \
 src/player_types.h src/files.h src/update.h src/tank.h src/physobj.h \
 src/virtobj.h src/floattext.h src/beam.h src/missile.h src/gameloop.h \
 src/clock.h src/client.h
