obj/network.o dep/network.d : src/network.cpp src/network.h src/player.h src/player_types.h \
 src/main.h src/debug.h src/globaltypes.h src/externs.h src/globaldata.h \
 src/text.h src/environment.h src/gfxData.h
