obj/teleport.o dep/teleport.d : src/teleport.cpp src/environment.h src/main.h src/debug.h \
 src/globaltypes.h src/externs.h src/globaldata.h src/text.h \
 src/network.h src/gfxData.h src/teleport.h src/virtobj.h src/tank.h \
 src/physobj.h src/floattext.h src/sound.h src/player.h \
 src/player_types.h
