obj/debris_pool.o dep/debris_pool.d : src/debris_pool.cpp src/debris_pool.h src/main.h \
 src/debug.h src/globaltypes.h src/externs.h src/globaldata.h src/text.h \
 src/environment.h src/network.h src/gfxData.h
