obj/floattext.o dep/floattext.d : src/floattext.cpp src/floattext.h src/main.h src/debug.h \
 src/globaltypes.h src/externs.h src/globaldata.h src/text.h \
 src/environment.h src/network.h src/gfxData.h src/virtobj.h
