obj/missile.o dep/missile.d : src/missile.cpp src/explosion.h src/main.h src/debug.h \
 src/globaltypes.h src/externs.h src/globaldata.h src/text.h \
 src/environment.h src/network.h src/gfxData.h src/physobj.h \
 src/virtobj.h src/missile.h src/decor.h src/debris_pool.h src/tank.h \
 src/floattext.h src/player.h src/player_types.h src/beam.h src/sound.h \
 src/aicore.h
