obj/sky.o dep/sky.d : src/sky.cpp src/externs.h src/globaldata.h src/main.h src/debug.h \
 src/globaltypes.h src/text.h src/environment.h src/network.h \
 src/gfxData.h src/sky.h src/files.h src/gameloop.h src/player.h \
 src/player_types.h
