obj/globaldata.o dep/globaldata.d : src/globaldata.cpp src/player.h src/player_types.h \
 src/main.h src/debug.h src/globaltypes.h src/externs.h src/globaldata.h \
 src/text.h src/environment.h src/network.h src/gfxData.h src/files.h \
 src/tank.h src/physobj.h src/virtobj.h src/floattext.h src/sound.h \
 src/debris_pool.h
