obj/client.o dep/client.d : src/client.cpp src/button.h src/main.h src/debug.h \
 src/globaltypes.h src/externs.h src/globaldata.h src/text.h \
 src/environment.h src/network.h src/gfxData.h src/files.h \
 src/satellite.h src/update.h src/client.h src/beam.h src/physobj.h \
 src/virtobj.h src/explosion.h src/missile.h src/teleport.h \
 src/floattext.h src/player.h src/player_types.h src/tank.h src/sky.h
