obj/explosion.o dep/explosion.d : src/explosion.cpp src/main.h src/debug.h src/globaltypes.h \
 src/externs.h src/globaldata.h src/text.h src/environment.h \
 src/network.h src/gfxData.h src/explosion.h src/physobj.h src/virtobj.h \
 src/missile.h src/decor.h src/debris_pool.h src/tank.h src/floattext.h \
 src/player.h src/player_types.h
