#pragma once
#ifndef ATANKS_SRC_SOUNDS_H_INCLUDED
#define ATANKS_SRC_SOUNDS_H_INCLUDED

/*
 * atanks - obliterate each other with oversize weapons
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

#include "externs.h"

/** @file sound.h
  * @brief declare centralized functions to play sounds.
**/

void play_explosion_sound(int32_t type,  int32_t x, int32_t vol, int32_t f_off);
void play_fire_sound     (int32_t type,  int32_t x, int32_t vol, int32_t f_off);
void play_music          ();
void play_natural_sound  (int32_t type,  int32_t x, int32_t vol, int32_t f_off);
void play_interface_sound(eSounds sound);

#endif // ATANKS_SRC_SOUNDS_H_INCLUDED

