obj/optionitemmenu.o dep/optionitemmenu.d : src/optionitemmenu.cpp src/optionitemmenu.h \
 src/optionitembase.h src/environment.h src/main.h src/debug.h \
 src/globaltypes.h src/externs.h src/globaldata.h src/text.h \
 src/network.h src/gfxData.h src/optiontypes.h src/menu.h \
 src/optionitem.h src/optionitemplayer.h src/button.h src/clock.h
