obj/player.o dep/player.d : src/player.cpp src/environment.h src/main.h src/debug.h \
 src/globaltypes.h src/externs.h src/globaldata.h src/text.h \
 src/network.h src/gfxData.h src/player.h src/player_types.h src/tank.h \
 src/physobj.h src/virtobj.h src/floattext.h src/menu.h src/optionitem.h \
 src/optionitembase.h src/optiontypes.h src/optionitemmenu.h \
 src/optionitemplayer.h src/button.h src/files.h src/missile.h \
 src/aicore.h
