obj/aicore.o dep/aicore.d : src/aicore.cpp src/aicore.h src/player_types.h src/main.h \
 src/debug.h src/globaltypes.h src/externs.h src/globaldata.h src/text.h \
 src/environment.h src/network.h src/gfxData.h src/floattext.h \
 src/virtobj.h src/player.h src/tank.h src/physobj.h src/missile.h \
 src/beam.h src/explosion.h
