obj/files.o dep/files.d : src/files.cpp src/main.h src/debug.h src/globaltypes.h \
 src/externs.h src/globaldata.h src/text.h src/environment.h \
 src/network.h src/gfxData.h src/player.h src/player_types.h src/files.h
