obj/gameloop.o dep/gameloop.d : src/gameloop.cpp src/main.h src/debug.h src/globaltypes.h \
 src/externs.h src/globaldata.h src/text.h src/environment.h \
 src/network.h src/gfxData.h src/files.h src/satellite.h src/update.h \
 src/land.h src/clock.h src/floattext.h src/virtobj.h src/tank.h \
 src/physobj.h src/explosion.h src/beam.h src/missile.h src/decor.h \
 src/debris_pool.h src/teleport.h src/sky.h src/sound.h src/gameloop.h \
 src/player.h src/player_types.h src/aicore.h src/shop.h
