obj/physobj.o dep/physobj.d : src/physobj.cpp src/physobj.h src/globaltypes.h src/virtobj.h \
 src/main.h src/debug.h src/externs.h src/globaldata.h src/text.h \
 src/environment.h src/network.h src/gfxData.h
