obj/optionitemcolour.o dep/optionitemcolour.d : src/optionitemcolour.cpp src/optionitemcolour.h \
 src/optionitembase.h src/environment.h src/main.h src/debug.h \
 src/globaltypes.h src/externs.h src/globaldata.h src/text.h \
 src/network.h src/gfxData.h src/optiontypes.h
