obj/shop.o dep/shop.d : src/shop.cpp src/shop.h src/player.h src/player_types.h \
 src/main.h src/debug.h src/globaltypes.h src/externs.h src/globaldata.h \
 src/text.h src/environment.h src/network.h src/gfxData.h src/files.h \
 src/gameloop.h
