#pragma once
#ifndef ATANKS_SRC_SHOP_H_INCLUDED
#define ATANKS_SRC_SHOP_H_INCLUDED

class LevelCreator; // From gameloop.h

// give people the chance to buy items
bool shop(LevelCreator* lvl_creator);


#endif // ATANKS_SRC_SHOP_H_INCLUDED

